from batou.component import Component
from batou.lib.file import File, Content
import ast
import batou.lib.git
import batou.lib.mercurial
import collections
import os.path
import pkg_resources
import urlparse


class Source(Component):

    dist_sources = '[]'
    sources = '[]'

    hg_hostfingerprints = {
        'bitbucket.org':
        '3f:d3:c5:17:23:3c:cd:f5:2d:17:76:06:93:7e:ee:97:42:21:14:aa',
        'code.gocept.com':
        '61:3f:8d:c9:b5:1f:65:79:68:d2:7d:1b:45:a2:c9:cb:b0:b4:32:de',
    }
    additional_hgrc_content = ''

    vcs_update = 'True'

    def _eval_attr(self, *names):
        for name in names:
            if isinstance(getattr(self, name), basestring):
                setattr(self, name, ast.literal_eval(getattr(self, name)))

    def configure(self):
        self._eval_attr('dist_sources', 'sources', 'vcs_update',
                        'hg_hostfingerprints')

        self.hg_hostfingerprints = collections.OrderedDict(
            x for x in sorted(self.hg_hostfingerprints.items()))

        self.provide('source', self)

        additional_hgrc = os.path.join(self.defdir, 'hgrc')
        if not self.additional_hgrc_content and os.path.exists(
                additional_hgrc):
            self |= Content('additional_hgrc', source=additional_hgrc)
            self.additional_hgrc_content = self._.content

        self.hgrc = File('~/.hgrc', source=pkg_resources.resource_filename(
            'batou_scm', 'resources/hgrc'))
        self += self.hgrc

        self.distributions = dict(
            self.add_clone(url) for url in self.dist_sources)
        self.clones = self.distributions.copy()
        self.clones.update(self.add_clone(url) for url in self.sources)

    VCS = {
        'hg': {
            'component': batou.lib.mercurial.Clone,
            'default-branch': 'default',
        },
        'git': {
            'component': batou.lib.git.Clone,
            'default-branch': 'master',
        },
    }

    def add_clone(self, url):
        parts = urlparse.urlsplit(url)
        vcs, _, scheme = parts.scheme.partition('+')
        if not vcs:
            raise ValueError(
                'Malformed VCS URL %r, need <vcs>+<protocol>://<url>, '
                'e.g. hg+ssh://hg@bitbucket.org/gocept/batou' % url)
        url = urlparse.urlunsplit((scheme,) + parts[1:])

        url, _, parameter_list = url.partition(' ')
        parameters = {}
        parameters['target'] = filter(bool, url.split('/'))[-1]
        parameters.update(x.split('=') for x in parameter_list.split())
        if 'branch' not in parameters and 'revision' not in parameters:
            parameters['branch'] = self.VCS[vcs]['default-branch']

        clone = self.VCS[vcs]['component']
        self += clone(url, vcs_update=self.vcs_update, **parameters)
        return parameters['target'], self._



from batou import UpdateNeeded
from batou.lib.buildout import Buildout
from batou.lib.file import Directory, File
import pkg_resources

try:
    from batou import SilentConfigurationError
except ImportError:
    # BBB batou < 1.1
    SilentConfigurationError = KeyError


class Buildout(Buildout):

    python = '2.7'
    version = '2.5.0'
    setuptools = '20.7'

    eggs_directory = '~/.batou-shared-eggs'
    versionpins = None

    def configure(self):
        try:
            self.source = self.require_one('source', self.host)
        except SilentConfigurationError:
            have_dists = False
            self.source = None
        else:
            have_dists = len(self.source.distributions)

        if have_dists:
            self.dist_names, distributions = zip(
                *sorted(self.source.distributions.items()))
            self.dist_paths = [clone.target for clone in distributions]
        else:
            self.dist_names = []
            self.dist_paths = []

        # A directory for eggs shared by buildouts within the deployment is
        # created for the service user. Assuming that a user cannot have
        # multiple non-sandboxed deployments and that development deployments
        # are sandboxed, eggs are never shared across deployments or users.
        self.eggs_directory = Directory(self.eggs_directory)
        self += self.eggs_directory

        self.overrides = File(
            'buildout_overrides.cfg',
            source=pkg_resources.resource_filename(
                'batou_scm', 'resources/buildout_overrides.cfg'))
        self.additional_config += (self.overrides,)

        super(Buildout, self).configure()

    @property
    def versions(self):
        result = {'setuptools': self.setuptools, 'zc.buildout': self.version}
        result.update({name: '' for name in self.dist_names})
        return sorted(result.items())

    __update_needed = None

    def verify(self):
        # We need to re-run buildout if any of the sources has changed.
        # This check incurs network access for each source checkout, so we
        # want to short-cut repeated calls.
        if self.source and self.__update_needed is None:
            try:
                self.source.assert_no_subcomponent_changes()
            except UpdateNeeded:
                self.__update_needed = True
            else:
                self.__update_needed = False
            for clone in self.source.clones.values():
                if clone.has_changes:
                    self.__update_needed = True
        if self.__update_needed:
            raise UpdateNeeded()
        super(Buildout, self).verify()


class BuildoutWithVersionPins(Buildout):
    """Buildout which expects a version attribute on the source component.

    It has to be a directory containing a file named `versions.cfg`.
    """

    def configure(self):
        source = self.require_one('source', self.host)
        self.versionpins = source.versions

        # XXX We should depend on all clones we are using in the buildout.cfg,
        # not just the versionpins.
        self.additional_config += (self.versionpins, )

        super(BuildoutWithVersionPins, self).configure()

    def verify(self):
        super(BuildoutWithVersionPins, self).verify()
        # XXX Only checking has_changes won't detect advanced SCM operations
        # like updating to an older revision, but as an 80-20 solution it's
        # quite good enough.
        if (self.versionpins.has_changes
                or self.versionpins.has_outgoing_changesets):
            raise UpdateNeeded()



from batou.conftest import root
from batou.lib.mercurial import Clone
from batou_scm.source import Source
import os.path
import pytest


@pytest.fixture
def source(root):
    source = Source(
        dist_sources=repr([
            'hg+https://example.com/foo',
            'hg+https://example.com/bar branch=BAR',
        ]),
        sources=repr([
            'hg+https://example.com/baz revision=BAZ',
        ]))
    root.component += source
    root.component.configure()
    return source


def test_all_clones_are_configured_as_subcomponents(source):
    clones = [s for s in source.sub_components if isinstance(s, Clone)]
    assert set(clones) == set(source.clones.values())
    assert set(clone.url for clone in clones) == set([
        'https://example.com/foo',
        'https://example.com/bar',
        'https://example.com/baz'])


def test_all_targets_are_derived_from_clone_urls(source):
    assert {clone.url: clone.target for clone in source.clones.values()} == {
        'https://example.com/foo': source.map('foo'),
        'https://example.com/bar': source.map('bar'),
        'https://example.com/baz': source.map('baz')}


def test_distributions_are_collected_separately(source):
    assert {
        clone.url: clone.target for clone in source.distributions.values()
    } == {
        'https://example.com/foo': source.map('foo'),
        'https://example.com/bar': source.map('bar'),
    }


def test_branches_are_read_from_parameters(source):
    assert source.clones['bar'].branch == 'BAR'


def test_branch_is_set_to_default_if_not_given(source):
    assert source.clones['foo'].branch == 'default'


def test_branch_is_not_set_if_revision_given(source):
    assert source.clones['baz'].branch is None


def test_revisions_are_read_from_parameters(source):
    assert source.clones['foo'].revision is None
    assert source.clones['bar'].revision is None
    assert source.clones['baz'].revision == 'BAZ'


def test_hg_hostfingerprints_are_sorted(source):
    source.hg_hostfingerprints['bitbucket.org'] = 'fpr1'
    source.hg_hostfingerprints['example.com'] = 'fpr2'
    source.configure()
    assert source.hgrc.content.startswith("""\
[hostfingerprints]
bitbucket.org = fpr1
code.gocept.com = 61:3f:8d:c9:b5:1f:65:79:68:d2:7d:1b:45:a2:c9:cb:b0:b4:32:de
example.com = fpr2
""")


def test_additional_hgrc_content_is_taken_from_file_if_present(source):
    open(os.path.join(source.defdir, 'hgrc'), 'w').write('foo')
    source.configure()
    assert 'foo' in source.hgrc.content


root  # XXX satisfy pyflakes



from batou import UpdateNeeded
from batou.conftest import root
from batou.lib.file import Directory
from batou_scm.buildout import Buildout, BuildoutWithVersionPins
from batou_scm.source import Source
import ConfigParser
import StringIO
import mock
import pytest


@pytest.fixture
def buildout(root):
    source = Source(
        dist_sources=repr([
            'hg+https://example.com/foo',
            'hg+https://example.com/bar',
        ]))
    root.component += source
    root.component += Buildout()
    root.component.configure()
    return root.component._


@pytest.fixture(scope='function')
def buildout_with_version_pins(root):
    source = Source(
        dist_sources=repr([
            'hg+https://example.com/foo',
            'hg+https://example.com/bar',
        ]),
        sources=repr(['hg+https://example.com/ver target=versions']))
    root.component += source
    source.versions = source.clones['versions']
    root.component += BuildoutWithVersionPins()
    root.component.configure()
    return root.component._


def read_config(content):
    config = ConfigParser.ConfigParser()
    config.readfp(StringIO.StringIO(content))
    return config


def test_buildout_doesnt_need_source_to_be_configured(root):
    root.component += Buildout()
    root.component.configure()
    assert '.batou-shared-eggs' in root.component._.overrides.content
    with pytest.raises(UpdateNeeded):
        root.component._.verify()


def test_dist_paths_are_listed_as_develop_paths_in_overrides(buildout):
    config = read_config(buildout.overrides.content)
    assert set([
        buildout.source.map('bar'),
        buildout.source.map('foo'),
    ]).issubset(config.get('buildout', 'develop +').split())


def test_no_dist_sources_configured_does_not_break(root):
    source = Source()
    root.component += source
    root.component += Buildout()
    root.component.configure()
    buildout = root.component._
    assert 'develop +=' not in buildout.overrides.content
    assert buildout.source is not None


def test_develop_paths_in_overrides_have_defined_order(buildout):
    config = read_config(buildout.overrides.content)
    develop_paths = config.get('buildout', 'develop +').split()
    assert develop_paths == sorted(develop_paths)


def test_dist_versions_are_unpinned_in_overrides(buildout):
    config = read_config(buildout.overrides.content)
    assert config.get('versions', 'foo') == ''
    assert config.get('versions', 'bar') == ''


def test_versions_of_setuptools_and_buildout_are_pinned_explicitly(buildout):
    config = read_config(buildout.overrides.content)
    assert config.get('versions', 'setuptools') == str(buildout.setuptools)
    assert config.get('versions', 'zc.buildout') == str(buildout.version)


def test_version_pins_in_overrides_have_defined_order(buildout):
    config = read_config(buildout.overrides.content)
    pins = config.items('versions')
    assert pins == sorted(pins)


def test_shared_egg_dir_is_created_for_service_user(buildout):
    config = read_config(buildout.overrides.content)
    eggs_dir = config.get('buildout', 'eggs-directory')
    assert eggs_dir == buildout.map('~/' + eggs_dir.split('/')[-1])
    assert any(component.path == eggs_dir
               for component in buildout.sub_components
               if isinstance(component, Directory))


def test_verify_checks_source(buildout):
    with mock.patch('batou.lib.buildout.Buildout.verify'):
        with mock.patch('batou.lib.mercurial.Clone.has_changes',
                        mock.PropertyMock()) as has_changes:
            has_changes.return_value = False
            buildout.source.assert_no_subcomponent_changes = mock.Mock()
            buildout.verify()
            assert buildout.source.assert_no_subcomponent_changes.called
            assert has_changes.called


def test_verify_result_is_cached(buildout):
    with mock.patch('batou.lib.buildout.Buildout.verify'):
        with mock.patch('batou.lib.mercurial.Clone.has_changes'):
            with mock.patch.object(
                    buildout.source, 'assert_no_subcomponent_changes') as ansc:
                ansc.side_effect = UpdateNeeded
                with pytest.raises(UpdateNeeded):
                    buildout.verify()
                with pytest.raises(UpdateNeeded):
                    buildout.verify()
                assert 1 == ansc.call_count


def test_buildout__BuildoutWithVersionPins__verify__1(
        buildout_with_version_pins):
    """`verify()` requests update on changed versions."""
    buildout = buildout_with_version_pins
    with mock.patch('batou.lib.mercurial.Clone.has_changes',
                    mock.PropertyMock()) as has_changes:
        has_changes.return_value = True
        with pytest.raises(UpdateNeeded):
            buildout.verify()


def test_buildout__BuildoutWithVersionPins__verify__2(
        buildout_with_version_pins):
    """`verify()` requests update on outgoing changes in versions."""
    buildout = buildout_with_version_pins
    with mock.patch('batou.lib.mercurial.Clone.has_outgoing_changesets',
                    mock.PropertyMock()) as has_outgoing_changesets, \
            mock.patch('batou.lib.mercurial.Clone.has_changes',
                       mock.PropertyMock()) as has_changes:
        has_changes.return_value = False
        has_outgoing_changesets.return_value = True
        with pytest.raises(UpdateNeeded):
            buildout.verify()


root  # XXX satisfy pyflakes



# This should be only one line. If it must be multi-line, indent the second
# line onwards to keep the PKG-INFO file format intact.
"""Batou components for managing source code in a development deployment.
"""

from setuptools import setup, find_packages
import glob
import os.path


def project_path(*names):
    return os.path.join(os.path.dirname(__file__), *names)


setup(
    name='batou_scm',
    version='0.3.2',

    install_requires=[
        'batou >= 1.0b26.dev0',
    ],

    extras_require={
        'test': [
            'batou >= 1.2',
            'mock',
        ],
    },

    entry_points={
        'console_scripts': [
            # 'binary-name = batou_scm.module:function'
        ],
    },

    author='gocept <mail@gocept.com>',
    author_email='mail@gocept.com',
    license='ZPL 2.1',
    url='https://bitbucket.org/gocept/batou_scm/',

    keywords='batou source code development checkout clone buildout',
    classifiers="""\
License :: OSI Approved :: Zope Public License
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 2 :: Only
"""[:-1].split('\n'),
    description=__doc__.strip(),
    long_description='\n\n'.join(open(project_path(name)).read() for name in (
        'README.txt',
        'HACKING.txt',
        'CHANGES.txt',
    )),

    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    data_files=[('', glob.glob(project_path('*.txt')))],
    zip_safe=False,
)
